;;;
;;;  The data list prototype
;;;
;;; $Revision: 1.2 $

(defproto datalist-proto '(type data name parents info missing 
                           missing-indicator) '() *object*)
	
;(send datalist-proto :slot-value 'missing-indicator *missing-indicator-default*)

(defmeth datalist-proto :isnew (type &key data name parents info 
       missing-indicator)
"Message args: (type &key data name parents)
Creates a data list with data and name specified.  The type is the 
name of a message that when called returns values and labels.  
parents are passed to the type.  The current choices are: 
 :variate :data data :name name   -- numeric variate
 :factor :data :name   -- factor with labels generated
 :factor1 :data :name  -- factor without a dropped level
 :effect :data :name   -- factor using effect coding with labels generated
 :interaction :parents (list x1 x2 ...) -- interaction of x1 x2..."
  (send self :missing-indicator missing-indicator)
  (send self :type type)
  (send self :data :data data :type type) 
  (send self :name name)
  (send self :parents parents)
  (send self :info info)
  self)

(defmeth datalist-proto :save ()
"Method args: ()
Returns lisp code that when funcalled recreates the datalist."
 `(send datalist-proto :new ,(send self :type)
        :data ',(send self :data) :name ',(send self :name)
        :type ',(send self :type) :info ',(send self :info)
        :missing-indicator ',(send self :missing-indicator)
        :parents (mapcar #'eval ',(mapcar #'(lambda (a) (send a :save))
                         (send self :parents)))))

(defmeth datalist-proto :type (&optional set) 
"Method args: (&optional set)
Sets/returns the type of the datalist.  The type begins with a colon like :variate
and specifies the name of the method that is used to compute the values to be returned
by the datalist."
  (when set (setf (slot-value 'type) set)) 
  (slot-value 'type))

(defmeth datalist-proto :parents (&optional set)
"Method args:  (&optional set)
Sets/returns the parents slot.  Used with interactions."
  (when set (setf (slot-value 'parents) set))(slot-value 'parents))

(defmeth datalist-proto :data (&key data type (set-missing (set-missing)))
"Method args (&key data type set-missing)
When data, the data slot is filled with the values in data.  The type of the
datalist is also set; since a variate must be all real, type :variate is
changed to :text if any non-missing elements of data are not numbers.  If not
set, this method returns the contents of the data slot with missing values
replaced by the contents of the keyword set-missing.  The usual choices are
not-a-number for numerics or 'missing for graphics."
  (when data  
   (let* ((missing (which (mapcar #'(lambda (s) (or (is-missing s)
                          (eq s (send self :missing-indicator)))) data)))
          (obs (reverse (set-difference (iseq (length data)) missing)))
          (isreal (every #'realp (select data obs))))
        (send self :type (cond ((and isreal (eq type :variate)) :variate)
                               ((eq type :variate) :text)
                               (t type))) 
        (setf (slot-value 'missing) missing)
        (cond ((eq (send self :type) :text)
               (setf (slot-value 'data) 
                     (mapcar #'(lambda (a) (if (numberp a) a
                                           (format nil "~a" a))) data)))
              (t (setf (slot-value 'data) data)))))
  (let* ((d (copy-seq (slot-value 'data))))
    (when (send self :missing) 
          (setf (select d (send self :missing)) set-missing))
    d))

(defmeth datalist-proto :missing () (slot-value 'missing))
(defmeth datalist-proto :observed () 
  (reverse (set-difference (iseq (send self :length))
                           (send self :missing))))
 

(defmeth datalist-proto :data-list (&rest args)
  (list (apply #'send self :data args)))

(defmeth datalist-proto :name (&optional new) 
"Method args: (&optional set)
Sets/returns the datalist's name."
  (when new (setf (slot-value 'name) new)) (slot-value 'name))

(defmeth datalist-proto :info (&optional new) 
"Method args: (&optional set)
Sets/returns the datalist's info slot."
  (when new (setf (slot-value 'info) new)) (slot-value 'info))

(defmeth datalist-proto :values (&rest args) 
"Method args: (&optional set)
Sets/returns the datalist's values."
  (apply #'send self (send self :type) :values args))

(defmeth datalist-proto :labels () (send self (send self :type) :labels))

(defmeth datalist-proto :missing-indicator (&optional set)
"Method args: (&optional set)
Sets/returns the datalist's missing value indicator."
   (if (null (slot-value 'missing-indicator))
       (setf (slot-value 'missing-indicator) *missing-indicator-default*))
	(when set (setf (slot-value 'missing-indicator) set))
	(slot-value 'missing-indicator))

(defmeth datalist-proto :variate (what &rest args)
"Method args: (what &rest args)
For a variate, if what = :values, returns (list data)
               if what = :labels, returns (list name)
               if what = :length, returns (length data)
               if what = :obs-length returns (length (send self :observed))"
  (case what
    (:values (list (apply #'send self :data args)))
    (:labels (list (send self :name)))
    (:length (length (send self :data)))
    (:obs-length (length (send self :observed)))))

(defmeth datalist-proto :case-numbers (&rest what) 
"Method args (&rest args)
Returns the same values as :variate."
  (apply #'send self :variate what))

(defmeth datalist-proto :text (&rest args) (apply #'send self :variate args))
(defmeth datalist-proto :temp (&rest args) (apply #'send self :variate args))

(defmeth datalist-proto :deleted (&rest args) (apply #'send self :variate args))

(defmeth datalist-proto :renamed (&rest what) (apply #'send self :variate what))

(defmeth datalist-proto :ones (&rest what) (apply #'send self :variate what))

(defmeth datalist-proto :bin-wts (&rest what) (apply #'send self :variate what))

(defmeth datalist-proto :mark-data (&rest what) 
  (apply #'send self :variate what))

(defmeth datalist-proto :factor (what &rest args)
  (defmeth self :info () "Factor--first level dropped")
  (case what
    (:values (make-factor (send self :data) ))
    (:labels (factor-levels (send self :data) :prefix (send self :name) ))
    (:length (length (first (send self :values))))
    (:obs-length (length (send self :observed)))))

(defmeth datalist-proto :effect (what &rest args)
  (defmeth self :info () "Factor--effects coding")
  (case what
    (:values (make-factor (send self :data) :effect t))
    (:labels (factor-levels (send self :data)  :drop-last t :drop-first nil
                      :prefix (send self :name) ))
    (:length (length (first (send self :values))))
    (:obs-length (length (send self :observed)))))

(defmeth datalist-proto :factor1 (what)
  (defmeth self :info () "Factor--all levels included")
  (case what
    (:values (make-factor (send self :data) :drop-first nil))
    (:labels (factor-levels (send self :data) :prefix (send self :name)
                     :drop-first nil))
    (:length (length (first (send self :values))))
    (:obs-length (length (send self :observed)))))

(defmeth datalist-proto :interaction (what)
  (case what
    (:values (apply #'rcross-terms 
              (mapcar #'(lambda (a) (send a :values)) (send self :parents))))
    (:labels (apply #'cross-names 
            (mapcar #'(lambda (a) (send a :labels)) (send self :parents))))
    (:length (send (first (send self :parents)) :length))
    (:obs-length (length (send self :observed)))))

(defmeth datalist-proto :length () 
  (send self (send self :type) :length))

(defmeth datalist-proto :obs-length ()
  (send self (send self :type) :obs-length))
; (send self (send self :type) :length))

(defun rcross-terms (x &rest args)
"Args: (x &rest args)
Arguments should be lists. Returns list of cross products, with the first
argument list varying slowest."
 (flet ((mmult (a b)
  (mapcar #'(lambda (a b) (if (and (numberp a) (numberp b)) (* a b) 0)) a b)))
  (case (length args)
    (0 (error "Too few arguments."))
    (1 (let ((y (first args)))
         (apply #'append 
           (mapcar #'(lambda (a) (mapcar #'(lambda (b) (mmult a b)) y)) x))))
    (t (rcross-terms x (apply #'rcross-terms args))))))

(defun term-name (x &rest args)
  (flet ((paste (x y) (format nil "~a*~a" x y)))
    (case (length args)
      (1 (paste x (first args)))
      (t (term-name x (apply #'term-name args))))))

(defun make-factor (x &rest args &key (effect nil) (test #'equal)
   (alphabetize *alphabetize-labels*))
"Args: (x &key (drop-first t) (test #'equal) (effect nil))
Returns a list of indicators sequences for the levels of X. TEST is
used to check equality of levels. If DROP-FIRST is true, the indicator
for the first level is dropped."
  (let* ((levels (remove-missing
                   (remove-duplicates (coerce x 'list) :test test)))
         (missing (find-missing x))
         (ans (if effect (apply #'make-factor-effect x 
                    (qsort1 levels :alphabetize alphabetize) args)
                   (apply #'make-factor1 x 
                          (qsort1 levels :alphabetize alphabetize) args))))
   (mapcar #'(lambda (a) (setf (select a missing) (select x missing))) ans)
   ans))

(defun qsort1 (x &rest args &key alphabetize)
"Function args: (x &key alphabetize)
If x is numeric, returns x ordered from smallest to largest.
If alphabetize is t sorts x using qsort.  Otherwise, returns x."
  (cond (alphabetize (apply #'qsort x args))
        ((null (position nil (mapcar #'numberp x))) (apply #'qsort x args))
        (t x)))

(defun make-factor1 (x levels &key (drop-first t) (test #'equal))
    (mapcar #'(lambda (lev) (if-else (map-elements test lev x) 1 0))
            (if drop-first (rest levels) levels)))

(defun make-factor-effect (x levels &key (test #'equal))
"Args: (x &key (test #'equal))
If x has p levels, returns a list of p-1 lists with list j equal to 1
if x is at level j, -1 if x is at level p, and 0 otherwise. TEST is
used to check equality of levels."
  (let* ((last-level (first (last levels))))
    (mapcar #'(lambda (lev) (if-else (map-elements test lev x) 1 
                           (if-else (map-elements test last-level x) -1 0)))
            (select levels (iseq (- (length levels) 1))))))

(defun factor-levels (x &key (prefix "") (drop-first t) (test #'equal)
                        missing-indicator (drop-last nil)
                        (alphabetize *alphabetize-labels*)
                        (square-brackets t)) 
"Args: (x &key (prefix \"\") (drop-first t))
Constructs name strings using unique levels in X and PREFIX."
  (let* ((levels (qsort1 (remove-duplicates 
                   (remove-missing (coerce x 'list)) :test test))))
    (mapcar #'(lambda (x) (if square-brackets
                             (format nil "~a[~a]" prefix x)
                             (format nil "~a~a" prefix x)))
            (cond (drop-first (rest levels))
                  (drop-last (reverse (rest (reverse levels))))
                  (t levels)))))

